/*-
 * Copyright (c) 2015 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Julian Wiesener <jw@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "vtstor.h"

/*
 * VRT_selecthttp is not public in Varnish <= 4.0.3.
 */
static struct http *
selecthttp(VRT_CTX, enum gethdr_e where)
{
	struct http *hp;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	switch (where) {
	case HDR_REQ:
		hp = ctx->http_req;
		break;
	case HDR_BEREQ:
		hp = ctx->http_bereq;
		break;
	case HDR_BERESP:
		hp = ctx->http_beresp;
		break;
	case HDR_RESP:
		hp = ctx->http_resp;
		break;
	default:
		WRONG("selecthttp 'where' invalid");
	}
	return (hp);
}

VCL_VOID __match_proto__(td_vmod_vtstor__init)
vmod_vtstor__init(VRT_CTX, struct vmod_vtstor_vtstor **vmip,
		  const char *vcl_name, VCL_INT nbuckets,
		  VCL_INT timeout, VCL_INT keymax, VCL_INT valmax)
{
	struct vmod_vtstor_vtstor *vmi;
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	(void)vcl_name;

	AN(vmip);
	AZ(*vmip);
	ALLOC_OBJ(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(vmi);
	vmi->hash = vtstor_hash_sha256;
	vmi->alloc = vtstor_alloc;
	vmi->dealloc = vtstor_dealloc;

#define CONFIG(f,d,v) do {			\
		vmi->f = (d);			\
		if ((v) > 0)			\
			vmi->f = (v);		\
	} while (0)

	CONFIG(buckets, DEFAULT_NBUCKETS, nbuckets);
	CONFIG(timeout, DEFAULT_TIMEOUT, timeout);
	CONFIG(key_max, DEFAULT_KEYMAX, keymax);
	CONFIG(value_max, DEFAULT_VALMAX,  valmax);
#undef CONFIG

	vmi->map = vtstor_init(vmi);
	*vmip = vmi;
	CHECK_OBJ_NOTNULL(*vmip, VMOD_VTSTOR_VTSTOR_MAGIC);
}

VCL_VOID __match_proto__(td_vmod_vtstor__fini)
vmod_vtstor__fini(struct vmod_vtstor_vtstor **vmip)
{
	AN(vmip);
	CHECK_OBJ_NOTNULL(*vmip, VMOD_VTSTOR_VTSTOR_MAGIC);
	vtstor_deleteall(*vmip);
	FREE_OBJ(*vmip);
	*vmip = NULL;
}

VCL_VOID __match_proto__(td_vmod_vtstor_store)
vmod_vtstor_store(VRT_CTX, struct vmod_vtstor_vtstor *vmi,
    VCL_STRING s, VCL_HEADER hdr)
{
	struct http *hp;
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(hdr);
	AN(hdr->what);

	hp = selecthttp(ctx, hdr->where);
	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);

	unsigned u, l = hdr->what[0];
	assert(l == strlen(hdr->what + 1));
	for (u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
		if(!strncasecmp(hp->hd[u].b, (hdr->what + 1), l))
			vtstor_store(vmi, (hp->hd[u].b + ++l), s);
	}
}

VCL_STRING __match_proto__(td_vmod_vtstor_get)
vmod_vtstor_get(VRT_CTX, struct vmod_vtstor_vtstor *vmi, VCL_HEADER hdr)
{
	struct http *hp;
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);

	AN(hdr);
	AN(hdr->what);
	hp = selecthttp(ctx, hdr->where);
	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);

	unsigned u, l = hdr->what[0];
	assert(l == strlen(hdr->what + 1));
	for (u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
		if(!strncasecmp(hp->hd[u].b, (hdr->what + 1), l))
			return(vtstor_get(vmi, (hp->hd[u].b + ++l)));
	}
	return(NULL);
}

VCL_VOID __match_proto__(td_vmod_vtstor_delete)
vmod_vtstor_delete(VRT_CTX, struct vmod_vtstor_vtstor *vmi, VCL_HEADER hdr)
{
	struct http *hp;
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);
	AN(hdr);
	AN(hdr->what);

	hp = selecthttp(ctx, hdr->where);
	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);

	unsigned u, l = hdr->what[0];
	assert(l == strlen(hdr->what + 1));
	for (u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
		if(!strncasecmp(hp->hd[u].b, (hdr->what + 1), l))
			vtstor_delete(vmi, (hp->hd[u].b + ++l));
	}
}

VCL_INT __match_proto__(td_vmod_vtstor_buckets)
vmod_vtstor_buckets(VRT_CTX, struct vmod_vtstor_vtstor *vmi)
{
	(void) ctx;
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);

	return vmi->buckets;
}

VCL_INT __match_proto__(td_vmod_vtstor_timeout)
vmod_vtstor_timeout(VRT_CTX, struct vmod_vtstor_vtstor *vmi)
{
	(void) ctx;
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);

	return vmi->timeout;
}

VCL_INT __match_proto__(td_vmod_vtstor_keymax)
vmod_vtstor_key_max(VRT_CTX, struct vmod_vtstor_vtstor *vmi)
{
	(void) ctx;
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);

	return vmi->key_max;
}

VCL_INT __match_proto__(td_vmod_vtstor_value_max)
vmod_vtstor_value_max(VRT_CTX, struct vmod_vtstor_vtstor *vmi)
{
	(void) ctx;
	CHECK_OBJ_NOTNULL(vmi, VMOD_VTSTOR_VTSTOR_MAGIC);

	return vmi->value_max;
}

VCL_STRING __match_proto__(td_vmod_vxid)
vmod_vxid(VRT_CTX)
{
	char *p;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
	CHECK_OBJ_NOTNULL(ctx->req->sp, SESS_MAGIC);

	p = WS_Printf(ctx->ws, "%u", ctx->req->sp->vxid);
	if (p == NULL)
		VSLb(ctx->vsl, SLT_VCL_Error,
		    "vmod vtstor: insufficient workspace in vxid()");
	return p;
}

VCL_STRING  __match_proto__(td_vmod_version)
vmod_version(VRT_CTX)
{
	return VERSION;
}

