/*-
 * Copyright (c) 2015 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Julian Wiesener <jw@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include "cache/cache.h"
#include "vrt.h"

#include "vcc_if.h"

#define DEFAULT_NBUCKETS 1024
#define DEFAULT_TIMEOUT 60
#define DEFAULT_KEYMAX 64
#define DEFAULT_VALMAX 4096

struct vmod_vtstor_vtstor;

typedef size_t (*malloc_func)(struct vmod_vtstor_vtstor *, size_t, char **);
typedef void (*dealloc_func)(struct vmod_vtstor_vtstor *, char **);
typedef uint32_t (*hash_func)(VCL_STRING);

struct vtstor_map;

struct vmod_vtstor_vtstor {
	unsigned                magic;
#define VMOD_VTSTOR_VTSTOR_MAGIC    0x5881bc28

	hash_func hash;
	malloc_func alloc;
	dealloc_func dealloc;
	size_t key_max;
	size_t value_max;

	unsigned buckets;
	unsigned timeout;
	struct vtstor_map *map;
};

uint32_t vtstor_hash_sha256(VCL_STRING);
size_t vtstor_alloc(struct vmod_vtstor_vtstor *, size_t, char **);
void vtstor_dealloc(struct vmod_vtstor_vtstor *, char **);

struct vtstor_map* vtstor_init(struct vmod_vtstor_vtstor *);
void vtstor_deleteall(struct vmod_vtstor_vtstor *);
void vtstor_store(struct vmod_vtstor_vtstor *,VCL_STRING, VCL_STRING);
VCL_STRING vtstor_get(struct vmod_vtstor_vtstor *, VCL_STRING);
void vtstor_delete(struct vmod_vtstor_vtstor *,VCL_STRING);
